"""
TODO:
 - metaclass:
  - cleanup confer
 - descriptors - check_type/validators don't handle setters lol
 - deep_frozen?
 - field:
  - frozen
  - pickle/transient
  - mangled
  - doc
  - derive
  - check_type
 - class
  - strict_eq
  - allow_setattr
  - mangler
 - observable
 - c/py gen
 - iterable
 - proto/jsonschema gen
 - enums
 - nodal
 - embedding? forward kwargs in general? or only for replace?
"""
